# jtextemacs.tcl - additional procedures for Emacs-like Text bindings
# 
# Copyright 1992-1994 by Jay Sekora.  All rights reserved, except 
# that this file may be freely redistributed in whole or in part 
# for nonprofit, noncommercial use.
# 
# 
# Essentially, call
#   j:tb:basic_bind Text    ---   for basic Text bindings
#   j:tb:emacs_bind Text    ---   for Emacs-like Text bindings
#   j:tb:vi_bind Text       ---   for vi-like Text bindings
# (the emacs_bind and vi_bind procedures include the basic_bind 
# procedures; you don't have to call both.)
# You can check $j_tb(modified,$WIDGET) to see if a text widget has been
# modified.  (Set it to false when you save the text widget.)

# TO DO:
# ^L
# sentence-manipulation stuff
# case change commands, transposition commands
# commands to do with mark?
# word deletion - fix to use buffer
# generalise movement to copying-to-cutbuffer and deletion
# IMPROVE ENTRY BINDINGS
# literal-insert for entry


######################################################################
# j:tkb:e:generic_arg - start generic argument
#   kind of clumsy: set repeat count to four, or multiply by four
######################################################################

proc j:tkb:e:generic_arg { W args } {
  global j_teb
  
  # set up prefix/repeat information if this widget hasn't been used yet
  if {! [info exists j_teb(prefix,$W)]} {
    set j_teb(prefix,$W) 0
  }
  if {! [info exists j_teb(repeat_count,$W)]} {
    set j_teb(prefix,$W) 1
  }

  if {$j_teb(prefix,$W) == 1 && $j_teb(repeat_count,$W) == 0} {
    set j_teb(repeat_count,$W) 16		;# ^U^U -> 4*4
    return
  }
  if {$j_teb(prefix,$W) == 0} {
    set j_teb(prefix,$W) 1
    set j_teb(repeat_count,$W) 0		;# special; -> 4 in repeatable
    return
  }
  set j_teb(repeat_count,$W) [expr {$j_teb(repeat_count,$W)*4}]
}

######################################################################
###  TEXT EMACS DELETION COMMANDS
######################################################################

# j:tkb:e:kill_line W K A - delete insert to end-of-line, setting cutbuffer
#   (arg handled by called procedure)
proc j:tkb:e:kill_line { W K A } {
  global j_teb
  set j_teb(modified,$W) 1
  
  # set up prefix/repeat information if this widget hasn't been used yet
  if {! [info exists j_teb(prefix,$W)]} {
    set j_teb(prefix,$W) 0
  }
  if {! [info exists j_teb(repeat_count,$W)]} {
    set j_teb(repeat_count,$W) 1
  }

  # Append to cutbuffer if previous command was line-kill; otherwise
  #   start with new cutbuffer:
  set my_name [lindex [info level 0] 0]
  if {! [string match $my_name $j_teb(last_command,$W)]} {
    set j_teb(cutbuffer) {}
  }
  
  # special-case prefix == 1 and repeat_count == 0 for Emacs ^U^U:
  #
  if {$j_teb(prefix,$W) == 1 && $j_teb(repeat_count,$W) == 0} {
    set j_teb(repeat_count,$W) 4
  }
  
  # if no argument, DON'T kill "\n" unless it's only thing at insert
  #
  if {$j_teb(repeat_count,$W) < 2} {
    j:tkb:clear_count $W			;# in case it's eg -1
    if {[$W index insert] == [$W index {insert lineend}]} then {
      append j_teb(cutbuffer) [$W get insert]
      $W delete insert
    } else {
      append j_teb(cutbuffer) [$W get insert {insert lineend}]
      $W delete insert {insert lineend}
    }
  } else {
    # with argument, kill that many lines (including "\n")
    j:tkb:repeatable {
      append j_teb(cutbuffer) [$W get insert {insert lineend + 1 char}]
      $W delete insert {insert lineend + 1 char}
    } $W
  }
  
  set j_teb(repeat_count,$W) 1
}

# j:tkb:e:kill_region W K A - delete selected region, setting cutbuffer
###   emacs region shouldn't be conflated with Text selection!
proc j:tkb:e:kill_region { W K A } {
  global j_teb
  set j_teb(modified,$W) 1

  j:tkb:clear_count $W

  set j_teb(cutbuffer) {}
  catch {
    set j_teb(cutbuffer) [$W get sel.first sel.last]
    $W delete sel.first sel.last
  }
}

# j:tkb:e:yank W K A - insert contents of cutbuffer
###   handling of argument needs changed---not count, but not ignored
proc j:tkb:e:yank { W K A } {
  global j_teb

  j:tkb:clear_count $W
  
  j:tb:insert_string $W $j_teb(cutbuffer)
}

######################################################################
###  TEXT EMACS MARK COMMANDS
######################################################################

# j:tkb:e:set_mark_command W K A - set emacs mark at current insert point
proc j:tkb:e:set_mark_command { W K A } {
  $W mark set emacs_mark insert
}

# j:tkb:e:exchange_point_and_mark W K A - swap insert point and emacs mark
proc j:tkb:e:exchange_point_and_mark { W K A } {
  if {[lsearch [$W mark names] emacs_mark] != -1} {
    set mark [$W index emacs_mark]
    $W mark set emacs_mark insert
    $W mark set insert $mark
    $W yview -pickplace insert
  } else {
    error "The mark is not set in text widget $W."
  }
}

