// *************************************************************************
//    BuiltinUserTupleRelation.C
// *************************************************************************

#include "BuiltinUserTupleRelation.h"

#include "TupleIterator.h"
#include "Globals.h"
#include "InitBuiltins.h"
#include "Debug.h"

// *************************************************************************
Cor_BuiltinUserTupleRelation::Cor_BuiltinUserTupleRelation(int arity,
    Cor_CArg *n, UserTupleProc user_op)
        : Cor_BuiltinRelation(arity, n, NULL, 0)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    op = user_op;
    Cor_InsertBuiltin(*n,this);
}

// *************************************************************************
Cor_BindEnv *Cor_BuiltinUserTupleRelation::get_next(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Tuple *tuple = get_next_tuple(iterator);

    if (tuple)    // Eliminate "unused" warning
        ;    // Nothing

    if (iterator.no_match())
        return NULL;
    else
        return iterator.bindenv;
}

// *************************************************************************
Cor_Tuple *Cor_BuiltinUserTupleRelation::get_next_tuple(
            Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_BindEnv *env = iterator.bindenv;

    if (iterator.arg_list.arity() != arity()) {
        iterator.set_no_match(); 
        return NULL;
    }

    if (iterator.ipos > 0) { 
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return NULL;
    }
    iterator.ipos++;
    iterator.stack_mark.get_mark();

    // first create the query tuple from the bindenv
    Cor_ArgList *alist = new Cor_ArgList(arity());
    Cor_BindEnv *queryenv = NULL;

    for (int i = 0; i < arity(); i++) {
        Cor_Term term;
        term.expr = iterator.arg_list[i];
        term.bindenv = env;
        //Cor_full_dereference_term(term);
        Cor_deep_dereference_term(term);
        if (!(((term.expr.is_var()) && (term.bindenv == env))
            || (term.isConstant())))
        {
            term.expr = iterator.arg_list[i];
                // Dereferencing lead to a
                // different bindenv -- cannot use dereferenced term.expr
                // with iterator.bindenv, so revert to original arg_list[i]
            queryenv = iterator.bindenv;            
        }
        (*alist)[i] = term.expr;
    }

    Cor_Tuple query(alist);
    query.instantiate(queryenv); 


    // now call the user routine to compute the (single) answer to the query
    Cor_Tuple *result = (*op)(&query); 

    if (result == NULL) {
        // no answer to query
        iterator.stack_mark.pop_to();
        iterator.set_no_match(); 
        return NULL;
    }

    if (Cor_unify_arg_lists(iterator.arg_list, result->args(), env, NULL)) {
        iterator.reset_no_match();
        return result;
    } else {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return NULL;
    }
}

// *************************************************************************

