//**************************************************************************
//    PvmCoralClient.C
//        A simple non-GUI Coral client using PVM
//**************************************************************************

#include <string.h>

#include "ClientProtocol.h"

    // This is a hack but otherwise entire Coral library is linked in
// #define  _NO_CORAL
// #include "../stream/PvmStream.C"
// #include "../stream/CoralStream.C"
#include "PvmStream.h"
#include <stdio.h>

#include "readline/readline.h" 
#include "readline/chardefs.h"
#include "readline/history.h"
#include "readline/keymaps.h"

//**************************************************************************

extern FILE  *rl_outstream;

extern "C" char *readline(char *);
extern "C" void add_history (char *);

//**************************************************************************
// trim_right() -
//    Remove spaces tabs and newlines from the right side, check to make
//    sure the command is terminated with a space, if everything looks
//    okay, return 1.
//
//    Return: 1 if trimmed command appears complete, 0 if incomplete
//
//**************************************************************************
int trim_right(char *str)
{
    int j;
    int len = strlen(str);
    char *spt;

    for (j=len-1, spt=str+j; j >= 0; j--, spt--){

        if ((*spt == ' ') || (*spt == '\t') || (*spt == '\n')){
            *spt = '\0';
            len--;
        } else 
            break;
    }

    if (len && (str[len-1] == '.'))
        return(1);
    else
        return(0);
}

//**************************************************************************
int main()
{
    char Prompt[100];
    char Buf[2000];
    char *ReadBuf;
    char PassWord[12];

    Cor_pvmbuf * pbuf = new Cor_pvmbuf();
    CoralStream * cs = new CoralStream(pbuf);
    
    int sync_cnt = 2;

    rl_outstream = stdout;

    strcpy(Prompt,"ready>> ");
    PassWord[0] = '\0';

    CoralStream *in  = cs;
    CoralStream *out = cs;

    if (!pbuf->Valid()) {
        cerr << "\nCOULD NOT GET CONNECTION TO CORAL SERVER\n";
        exit(1);
    }

    // send empty command to server to find out if server is willing to
    // accept this connection.
    
    out->puts(".");
    out->printf("\nsync(%d). yield.\n",sync_cnt);
    out->flush();

    int done = 0;
    int sync = 0;
    int ok_cnt = 2;

    do {
        if (!(in->gets(Buf, 1000))) 
            break;

        if (Buf[4] == ':'){

            if (strncmp(Buf,CP_OKAY,5) == 0) {
                if (sync)
                    if (--ok_cnt == 0)
                        done = 1;

            } else if (strncmp(Buf,CP_SYNC,5) == 0) {
                int val;
                sscanf(&(Buf[5]), "%d", &val);
                if (val == sync_cnt){
                    sync = 1;
                    sync_cnt++;
                }

            } else if (strncmp(Buf,CP_CERR,5) == 0) {
                cout << (&(Buf[5]));

            } else if (strncmp(Buf,CP_MESG,5) == 0) {
                cout << (&(Buf[5]));

            } else if (strncmp(Buf,CP_LOGIN,5) == 0) {
            } else if (strncmp(Buf,CP_USER,5) == 0) {
                cout << Buf;

            } else if (strncmp(Buf,CP_PSWD,5) == 0) {
                cout << Buf;
                // Save the password for use with Explain
                strncpy(PassWord,&(Buf[5]),8);
                PassWord[8] = '\0';
                int i = strlen(PassWord);
                if (PassWord[i-1] == '\n')
                    PassWord[i-1] = '\0';

            } else if (strncmp(Buf,CP_R_ID,5) == 0) {
                cout << Buf;

            } else if (strncmp(Buf,CP_PRMT,5) == 0) {

                strcpy(Prompt,&(Buf[5]));
                Prompt[strlen(Prompt)-1] = ' ';

            } else if (strncmp(Buf,CP_EXIT,5) == 0) {

                cout << "exiting.\n";
                return(0);

            } else {
                cerr << "Warning, unimplemented protocol header.\n";
                cout << Buf;
            }
        } else {
            cout << Buf;
        }

    } while(!done);
    cerr << "Connected to Coral PVM server.\n";
    
    // now start to accept "real" command

    Buf[0] = '\0';
 
    while (1) {

        if (*Buf){
            if (!(ReadBuf = readline("")))
                break;
        } else
            if (!(ReadBuf = readline(Prompt)))
                break;

        if (trim_right(ReadBuf)){

            strcat(Buf,ReadBuf);
            free(ReadBuf);

        } else {
            strcat(Buf,ReadBuf);
            free(ReadBuf);
            continue;
        }

        if (*Buf) 
          add_history(Buf);

        out->puts(Buf);
        out->printf("\nsync(%d). yield.\n",sync_cnt);
        out->flush();

        int done = 0;
        int sync = 0;
        int ok_cnt = 2;

        do {
            if (!(in->gets(Buf, 1000))) 
                break;

          if (Buf[4] == ':'){

              if (strncmp(Buf,CP_OKAY,5) == 0) {
                if (sync)
                    if (--ok_cnt == 0)
                        done = 1;

              } else if (strncmp(Buf,CP_SYNC,5) == 0) {
                int val;
                sscanf(&(Buf[5]), "%d", &val);
                if (val == sync_cnt){
                    sync = 1;
                    sync_cnt++;
                }

              } else if (strncmp(Buf,CP_CERR,5) == 0) {
                 cout << (&(Buf[5]));

              } else if (strncmp(Buf,CP_MESG,5) == 0) {
                 cout << (&(Buf[5]));

              } else if (strncmp(Buf,CP_LOGIN,5) == 0) {
              } else if (strncmp(Buf,CP_USER,5) == 0) {
                 cout << Buf;

              } else if (strncmp(Buf,CP_PSWD,5) == 0) {
                 cout << Buf;
                  // Save the password for use with Explain
                strncpy(PassWord,&(Buf[5]),8);
                PassWord[8] = '\0';
                int i = strlen(PassWord);
                if (PassWord[i-1] == '\n')
                    PassWord[i-1] = '\0';

              } else if (strncmp(Buf,CP_R_ID,5) == 0) {
                 cout << Buf;

              } else if (strncmp(Buf,CP_PRMT,5) == 0) {

                strcpy(Prompt,&(Buf[5]));
                Prompt[strlen(Prompt)-1] = ' ';

              } else if (strncmp(Buf,CP_EXIT,5) == 0) {

                 cout << "exiting.\n";
                return(0);

              } else {
                 cerr << "Warning, unimplemented protocol header.\n";
                 cout << Buf;
              }
          } else {
             cout << Buf;
          }

        } while(!done);

        Buf[0] = '\0';
    }

    out->puts("\nquit.\n");
    out->flush();

    delete cs;

    cerr << "\nbye!\n";
    return(0);
}
//**************************************************************************

