#ifdef ASYN_SOCK_REL

// *************************************************************************
//    AsynSockRelation.C
// *************************************************************************

#include "AsynSockRelation.h"

#include "Util.h"
#include "Globals.h"
#include "CoralError.h"
#include "Unify.h"
#include "Debug.h"
#include "TupleIterator.h"
#include "ClientProtocol.h"

// *************************************************************************
Cor_AsynSockRelation::Cor_AsynSockRelation(char *ln, int arity, char *host,
     int port, char *ws, char *rn)
    : Cor_DerivedRelation(Cor_CArg(ln), arity) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    _delta_done = 0; 
    
    strcpy(_local_name,ln);
    _arity = arity;
    strcpy(_host,host);
    _port = port;
    strcpy(_work_space,ws);
    strcpy(_relation_name, rn);

    // turn off the checking for subsumption.it's done on remote server
    check_subsum = 0;
}

// *************************************************************************
Cor_AsynSockRelation::~Cor_AsynSockRelation()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
}

CoralStream *Cor_AsynSockRelation::setup()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    CoralStream *cs;
    Cor_sockbuf * sbuf = new Cor_sockbuf(_host);

    int valid = sbuf->valid();
    if (!valid) {
        *(Cor_exEnv->err_strm)
           << "Error: could not get connected with coral server "
           << _host << endl;
        return NULL;
    }
    else {
        *(Cor_exEnv->err_strm) << "Connected to coral server "
                               << _host << endl;
        cs = new CoralStream(sbuf);
        
        // switch to the correct work space
        *cs << "clear(interactive_mode).\n";
        *cs << "clear(verbose_answer).\n";
        *cs << "assign(answer_style,3).\n";
        cs->flush();

        *cs << "ws_create(\"" << _work_space << "\").";
        *cs << "ws_current(\"" << _work_space << "\").";
        *cs << "yield.\n";
        cs->flush();

        return cs;
    }
}

//*************************************************************************
Cor_Tuple *Cor_AsynSockRelation::get_next_tuple(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *(Cor_exEnv->err_strm)
        << "Error: should not call get_next_tuple for asyn_sock_relation "
        << name << endl;
    *(Cor_exEnv->err_strm)
        << "       Check the module to see if magic rewriting is used and "
        << "dumb rule is added for asyn_sock_relation.\n";
    iterator.set_no_match();
    return NULL;
}

// *************************************************************************
Cor_BindEnv *Cor_AsynSockRelation::get_next(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *(Cor_exEnv->err_strm)
        << "Error: should not call get_next for asyn_sock_relation "
        << name << endl;
    iterator.set_no_match();
    return NULL;
}

// *************************************************************************
int Cor_AsynSockRelation::insert_tuple(Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    CORAL_error(COR_BAD_TUPLE_INSERT, name.symbol(), "insert_tuple");
    return -1;
}

// *************************************************************************
void Cor_AsynSockRelation::print_name(ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *out << "local_name = " << _local_name << endl;
    *out << "arity = " << _arity << endl;
    *out << "host = " << _host << endl;
    *out << "port = " << _port << endl;
    *out << "work_space = " << _work_space << endl;
    *out << "relation_name = " << _relation_name << endl;
}

// *************************************************************************
#endif // ASYN_SOCK_REL

