// *************************************************************************
//    Relation.C
// *************************************************************************

#include <stdio.h>
#include <stdlib.h>

#ifdef ASYN_SOCK_REL

#include <strstream.h>

#include "QFMaterializedModuleData.h"
#include "AsynSockRelation.h"

#endif

#include "Relation.h"

#include "Globals.h"
#include "Interface.h"
#include "RelationMisc.h"
#include "StackMark.h"
#include "TupleIterator.h"
#include "BuiltinSymbols.h"
#include "HashSimpleRelation.h"
#include "BindEnv.h"
#include "CoralError.h"
#include "Unify.h"
#include "Execute.h"
#include "Debug.h"


// *************************************************************************

int Cor_C_cont_id = 1;            // Used for ReturnUnification

// *************************************************************************
Cor_Relation::Cor_Relation()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char namep[100];
    sprintf(namep,"Rel_%d", Cor_RelCount++);
    name = Cor_CArg(namep);
    _arity         = 0;

    local_options  = 0;
    count          = 0;
    scan_count     = 0;
    ref_count      = 1;
    carg_count     = 0;
    single_success = 0;
    has_release    = 1;
    _borrow        = 0;
    pl_local       = 0;
    
#ifdef ASYN_SOCK_REL
    _delta_done    = 1;
    first_send_delta = 1;
    open_queries = 0;
#endif
    
    check_subsum = Cor_exEnv->C_check_subsum_default;

    r_kind = COR_R_ANSWER;  // default

    delta_indexed = Cor_exEnv->C_index_deltas_default;

    ordsearch     = 0;
    for_ordsearch = NULL;
    mp_done_reln  = NULL;
    agg_sel_info  = NULL;

    collector     = NULL;
}

// *************************************************************************
Cor_Relation::Cor_Relation(int new_arity, int delta)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char namep[100];
    sprintf(namep,"Rel_%d", Cor_RelCount++);
    name = Cor_CArg(namep);
    _arity = new_arity;

    local_options = 0;
    count = 0;
    scan_count = 0;
    ref_count  = 1;
    carg_count = 0;

    single_success = 0;
    check_subsum = Cor_exEnv->C_check_subsum_default;
    has_release    = 1;
    _borrow        = 0;
    pl_local       = 0;
    
#ifdef ASYN_SOCK_REL
    _delta_done    = 1;
    first_send_delta = 1;
    open_queries = 0;
#endif
    
    r_kind = COR_R_ANSWER;    // default

    if (delta == -1)
        delta_indexed = Cor_exEnv->C_index_deltas_default;
    else
        delta_indexed = delta;

    agg_sel_info = NULL;

    _hash  = UnknownHashValue;
    _eq_id = UnknownHashValue;

    ordsearch = 0;
    for_ordsearch = NULL;
    mp_done_reln  = NULL;

    collector     = NULL;
}

// *************************************************************************
Cor_Relation::~Cor_Relation()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    PURE_PRINTF
    purify_printf("Delete of relation %p\n", this);
#endif

    delete for_ordsearch;
    delete mp_done_reln;

    Cor_AggSelInfo *cur, *prev = NULL;

    for (cur = agg_sel_info; cur; cur = cur->next) {
        delete prev;
        prev = cur;
    }
    delete prev;
    agg_sel_info = NULL;
}

// *************************************************************************
void Cor_Relation::print(Cor_BindEnv *env, ostream *file, char *) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    this->print(env, file);
}

// *************************************************************************
void Cor_Relation::print(Cor_BindEnv*, ostream *file) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *file << "{";
    printon(file);
    *file << "}";
}

// *************************************************************************
void Cor_Relation::sprint(char *str, int *pos, Cor_BindEnv *env) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    sprintf(str + *pos, "{");
    (*pos)++;

    Cor_StackMark   stack_mark0;
    Cor_BindEnv     env0(arity());
    Cor_ArgList    *args = make_vararglist(arity());
    Cor_TupleIterator    iter((Cor_Relation*)this, *args, &env0);
    Cor_StackMark   stack_mark1;
    Cor_Term        term1;

    Cor_BindEnv *env1 = iter.get_next();

    if    (!(iter.no_match())) {
        for (int iarg = 0; iarg < arity(); iarg++) {
            if (iarg > 0) {
                sprintf(str+ *pos, ", ");
                (*pos) += 2;
            }
            term1 = env1->lookup(iarg);

            if (!term1.expr.is_empty())
                term1.expr.sprint(str,pos,env);
        }

        for (;;) {
            Cor_BindEnv *env1 = iter.get_next();
            if (iter.no_match()) break;
            sprintf(str + *pos, ",");
            (*pos)++;
            for (int iarg = 0; iarg < arity(); iarg++) {
                if (iarg > 0) {
                    sprintf(str+ *pos, ", ");
                    (*pos) += 2;
                }
                term1 = env1->lookup(iarg);

                if (!term1.expr.is_empty())
                    term1.expr.sprint(str, pos, env);
            }
            stack_mark1.pop_to();
        }
    }

    stack_mark0.pop_to();
    delete args;

    sprintf(str    + *pos, "}");
    (*pos)++;
}

// *************************************************************************
void Cor_Relation::print_name( ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *out << name.symbol() << "/" << arity() << "\t : (generic)\n";
}

// *************************************************************************
//  Inserts a tuple into the relation.
//  Checks for subsumption depending on this->check_subsum.
//
//  If dont_rename is set, it does not perform renaming,
//  else it renames (args, env);  variables in dont_rename_env
//  are not renamed.
//  The (renamed) (args,env) is stored in (copy_args, copy_env)
//
//  If OrderedSearch is used, special things are done, and
//  (copy_args, copy_env) (with additional fields) is inserted into context.
//  Otherwise (copy_args, copy_env) is inserted into the relation.
//
//  Note: Variables in dont_rename_env are not renamed
// *************************************************************************
int Cor_StorageRelation::insert_new(Cor_ArgList& args, Cor_BindEnv *tupenv)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

#ifdef    INSERT_VERBOSE
    cout << "Entering function: " << __FUNCTION__ << endl;
    cout << "Attempt to insert: " << args << endl;
    cout << "Instan to: ";
    args.print(tupenv,&cout);
    cout << endl;
#endif

    Cor_BindEnv *env = NULL;

    if (Cor_exEnv->C_interrupt_raised)
        return(0);

#    ifdef DO_TRACE
        if (DISPLAY_INSERTIONS) {
            *(Cor_exEnv->trace_strm) << 
                "[Starting to insert " << name.symbol() << "(";
            args.print(tupenv, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << ")";
        }
#    endif

    if (args.arity() != arity()) {
        *(Cor_exEnv->trace_strm) << 
            "Relation::insert_new - attempt to insert tuple of arity "
            << args.arity() << endl
            << "  - in relation " << name.symbol() << " of arity "
            << arity() << endl;
        return(0);
    }

    if (args.has_grouping_args())
        return(insert_grouping(args,tupenv));

    // This can only happen with non-ground facts? -SPF
#ifdef    SPF_DEBUG
    if (tupenv == NULL) {
        int env_size = args.max_vars();

        if (env_size != 0) {
            // WARNING - this space is never reclaimed.  Make sure insert_new
            // is not called with a null bindenv

            //env = new Cor_BindEnv(args.max_vars());
            // Attempted to insert a non-ground tuple?
            cout << "env_size != 0\n";
            return(0); 
        } else {
            env = NULL;
        }
    } else {
        env = tupenv;
    }
#else
    env = tupenv;
#endif

    // first some subsumption checks
    if (check_subsum == 1) {

        Cor_Tuple *newtuple;

        // If non-ground facts are present, a new tuple needs to be created
        // before performing subsumption checks, otherwise the old tuple
        // itself can be used.

        Cor_ArgList *args2 = args.dup();
        newtuple = new Cor_Tuple(args2);
        newtuple->instantiate(env);

        if (!ordsearch) {
            // ordinary case
            int stat = is_subsumed(newtuple);
            if (stat) {
#                ifdef DO_TRACE
                    if (DISPLAY_INSERTIONS) {
                        *(Cor_exEnv->trace_strm) << " - subsumed]\n";
                    }
#                endif
                delete newtuple->_args;
                delete newtuple;
                return(stat < 0);
            }
/*
            if (is_subsumed(newtuple)) {
#                ifdef DO_TRACE
                    if (DISPLAY_INSERTIONS) {
                        *(Cor_exEnv->trace_strm) << " - subsumed]\n";
                    }
#                endif
                delete newtuple->_args;
                delete newtuple;
                return(0);
            }
*/

        } else {
            // OrdSearch: these are just preliminary checks
            // subsumption against Context done later

            if (r_kind == COR_R_MAGIC) {
                // subsumption checking is first done against the mp_done_reln
                if (mp_done_reln && mp_done_reln->is_subsumed(newtuple)) {
#                    ifdef DO_TRACE
                        if (DISPLAY_INSERTIONS) {
                            *(Cor_exEnv->trace_strm) << " - subsumed]\n";
                            *(Cor_exEnv->trace_strm) << 
                                "        OrdSearch: using mp_done\n";
                        }
#                    endif
                    delete newtuple->_args;
                    delete newtuple;
                    return(0);
                }
            } else if (r_kind != COR_R_SUPPLEMENTARY) {
                // don't want any subsumption checking done if
                // (ordsearch && (r_kind == COR_R_SUPPLEMENTARY))

                if (is_subsumed(newtuple)) {
#                    ifdef DO_TRACE
                        if (DISPLAY_INSERTIONS) {
                            *(Cor_exEnv->trace_strm) << " - subsumed]\n";
                        }
#                    endif
                    delete newtuple->_args;
                    delete newtuple;
                    return(0);
                }
            }
        }
        delete newtuple->_args;
        delete newtuple;
    }

    // now add the tuple to the appropriate relation

    Cor_GoalNode *g_node;
    int negated_g_node = 0;
    Cor_Table *F;
    Cor_ClauseNames *c_names;
    int j, num_cnme;

    if (ordsearch
        && ((r_kind == COR_R_MAGIC) || (r_kind == COR_R_SUPPLEMENTARY))
        && (Cor_T_Stack->count > 0))
    {
        F = Cor_T_Stack->top();
        if (F && F->moduledata && F->moduledata->curr_rInfo &&
            F->moduledata->curr_rInfo->rule)
        {
            c_names = F->moduledata->curr_rInfo->rule->clause_names;
            num_cnme = c_names->count();
            for (j=0; j < num_cnme; j++) {
                if ( (negated_g_node = c_names->ith_neg(j)) )
                    break;
            }
        }
    }

    // For Ordered Search, the m_p and sup_ facts in Context are maintained
    // as auxiliary relations "for_ordsearch" associated with the regular
    // m_p and sup_ relations. The "for_ordsearch" relation is used as
    // an index into context.
    if (ordsearch && ((r_kind == COR_R_MAGIC) ||
        (r_kind == COR_R_SUPPLEMENTARY)) && (!for_ordsearch))
    {
        for_ordsearch = new Cor_HashSimpleRelation(arity());
        // What other indices besides the default indices need to be
        // created?
        // WARNING: if linked relations are to be used, this may have to
        // be modified
    }

    //  divesh: added stuff for optimizing insertion of facts by
    //  insert_new.    when there is a for_supp_index relation, then
    //  this is a magic relation, and a previous call to goal_id
    //  should have inserted this same fact in the for_supp_index
    //  relation.    hence, instead of CopyArgs and deleting the
    //  fact there, we just "move" the fact from the for_supp_index
    //  relation to the main relation.

    // WARNING::  ReturnUnify optimizations needed here!!!

    Cor_ArgList *copy_args;
    Cor_BindEnv *copy_env = NULL;
    int env_size;

        // there is no auxiliary relation and CopyArgs needs to be called
#if    1
    // Only need something like this for non-ground?
    copy_args = new Cor_ArgList(arity());
    env_size = Cor_CopyArgs(copy_args, &args, env,
                arity(),(Cor_BindEnv*)NULL);
#else
    copy_args = &args;
     // Need to check and set this
    env_size = 0;
#endif

    if (env_size != 0) {
        *(Cor_exEnv->out_strm)
         << "\nWarning: Attempt to insert a non-ground fact into relation "
            << name.symbol() << ".\n(the insert was ignored).\n";
#ifdef    DEBUG
        cout << "Insert of: " << *copy_args << endl;
#endif
        delete copy_args;
        return(0);
    }

    copy_env = NULL;


#        ifdef DO_TRACE
            if (DISPLAY_NEW_INSERTIONS && !DISPLAY_INSERTIONS ) {
                *(Cor_exEnv->trace_strm) << 
                    "[Inserting " <<  name.symbol() << "(";
                copy_args->print(copy_env, Cor_exEnv->trace_strm);
                *(Cor_exEnv->trace_strm) << ")";
            }
#        endif

        if (ordsearch
            && ((r_kind == COR_R_MAGIC) || (r_kind == COR_R_SUPPLEMENTARY))) {

            g_node = new Cor_GoalNode(copy_args,copy_env,this,negated_g_node);

        } else {
            // do the regular stuff and return 1
            Cor_Tuple *tuple1;
            tuple1 = new Cor_Tuple(copy_args);
            tuple1->instantiate(copy_env);

            // WARNING:: Will need to eventually create a version of the env
            // and bump up its size as required.

            tuple1->contid = Cor_C_cont_id++; // See below.

            if (!insert_tuple(tuple1)) {
                delete tuple1;
            }

#            ifdef DO_TRACE
                if (DISPLAY_INSERTIONS | DISPLAY_NEW_INSERTIONS) {
                    if (env_size == 0)
                        *(Cor_exEnv->trace_strm) << " - done (ground)]\n";
                    else
                        *(Cor_exEnv->trace_strm) << 
                            " - done (env-size:" <<  env_size << ")]\n";
                }
#            endif

            return(1);
        }

    // finally, for ordered search relations a special kind of
    // subsumption needs to be done on the basis of the position of
    // insertion of g_node in context

    // for the ordinary case, the tuple has been inserted previously
    // the program control reaches this point only with OrdSearch

    ASSERT((ordsearch && ((r_kind == COR_R_MAGIC) ||
        (r_kind == COR_R_SUPPLEMENTARY))));

    // For ReturnUnification; see comments above
    g_node->contid = Cor_C_cont_id++;

    int stat = insert_in_context(g_node, (Cor_GoalNode *) NULL);

    if (!stat) {
        if (! g_node->has_no_args()) {
            delete g_node->_args;
        }
        delete g_node;
    }

#    ifdef DO_TRACE
        if (DISPLAY_INSERTIONS | DISPLAY_NEW_INSERTIONS) {
            if (env_size == 0)
                *(Cor_exEnv->trace_strm) << " - done (ground)]\n";
            else
                *(Cor_exEnv->trace_strm) << 
                    " - done (env-size:" <<  env_size << ")]\n";
        }
#    endif
    return(stat);
}

// *************************************************************************
void Cor_Relation::printon(ostream *file, char *) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    this->printon(file);
}

// *************************************************************************
void Cor_Relation::printon(ostream *file) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StackMark stack_mark0;
    Cor_BindEnv env0(arity());
    Cor_ArgList *args = make_vararglist(arity());
    Cor_TupleIterator iter((Cor_Relation*)this, *args, &env0);
    Cor_StackMark stack_mark1;

    Cor_BindEnv *env1 = iter.get_next();
    if (!(iter.no_match())) {
        for (int iarg = 0; iarg < arity(); iarg++) {
            if (iarg > 0)
                *file << ", ";
            env1->lookup(iarg).printon(file);
        }

        for (;;) {
            Cor_BindEnv *env1 = iter.get_next();
            if (iter.no_match()) break;
            *file << ",";
            for (int iarg = 0; iarg < arity(); iarg++) {
                if (iarg > 0)
                    *file << ", ";
            env1->lookup(iarg).printon(file);
            }
            stack_mark1.pop_to();
        }
    }
    stack_mark0.pop_to();
    delete args;
}

// *************************************************************************
void Cor_Relation::print_facts(ostream *file, Cor_ArgList *arglist,
        int as_table ) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StackMark stack_mark0;
    Cor_ArgList *args;

    if (!arglist)
        args = make_vararglist(arity());
    else {
        if (arglist->arity() != arity()) {
            CORAL_error(COR_BAD_ARGLIST, NULL, "print_facts");
            *(Cor_exEnv->trace_strm) << "Bad arity "
                << arglist->arity() << " - "
                <<  arity() << " expected\n";
            return;
        }
        args = arglist;
    }

    Cor_BindEnv env0(arity());
    Cor_TupleIterator iter((Cor_Relation*)this, *args, &env0);
    Cor_StackMark stack_mark1;
    Cor_Term term;
    for (;;) {
        Cor_BindEnv *env1 = iter.get_next();
        if (env1) // Eliminate "unused" warning
            ;    // Nothing

        if (iter.no_match())
            break;
        if (! as_table)
            *file << name.symbol() << "(";

        for (int iarg = 0; iarg < arity(); iarg++) {
            if (iarg > 0) {
                if (as_table)
                    *file << " ";
                else
                    *file << ",";
            }
            term.expr = (*args)[iarg];
            term.bindenv = &env0;
            term = term.dereference();
            term.printon(file);
        }

        if (as_table)
            *file << endl;
        else
            *file << ").\n";
        stack_mark1.pop_to();
    }

    stack_mark0.pop_to();
    if (!arglist)
        delete args;
}

// *************************************************************************
//  Hashing for nested sets
//
//  We use two fields for comparing sets:
//  HashVal Relation::_hash, and
//  HashVal Relation::_eq_id
//
//  (1) _hash is computed using HashVal Relation::hash(BindEnv *)
//      which computes a hash value for a relation based on the tuples
//      in the relation.  It uses a commutative, associative hashing
//      function (multiplication), as a result of which the hash value
//      of a relation does not depend on the order in which tuples are
//      accessed for computing the hash value.
//      This is used as a sufficient condition for INEQUALITY, i.e.,
//      if two hash values are distinct, so are the relations.
//      However, as with all hash values, two relations could be distinct
//      and have the same hash value.
//
//  (2) _eq_id is an "id" field which is used as a sufficient
//      condition for equality, i.e. if two _eq_id fields are the
//      same, the relations are the same. However, it may be the case that
//      _eq_id of one relation is different from the _eq_id field of
//      another relation, while the two relations are indeed equal.
//      In which case, we physically compare the two relations, and
//      if equal, set the _eq_id field of each to the lower of the two.
//
// *************************************************************************
//    compute the hash value of a relation argument.

HashVal Cor_Relation::hash(Cor_BindEnv * /* env is ignored */)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    HashVal h_accum = 1;
    HashVal h_temp;
    Cor_Tuple *T;

    if (_hash != UnknownHashValue) {
        // then set has been hashed, so
        return (_hash);
    }

    FOR_EACH_TUPLE(T, this) {
        h_temp = T->hash();     //  hash the element T

        if (h_temp == VarHashValue) {
            h_accum = VarHashValue;
            break;
        }
        else {
            h_accum = h_accum * h_temp;
        }
    } END_EACH_TUPLE(T)

    _hash = h_accum;
    return(_hash);
}

// *************************************************************************
/*
    Cardinality of a tuple in a relation.
    INPUT: a pointer to a tuple, T
    OUTPUT: the number of occurrences of (*T) in (*this)
    NOTE:  We assume that T is ground, and all tuples of *this are ground.
    NOTE:  this could be made more efficient, by iterating only through the
        hash bucket for T, instead of all of the rel, should large sets
        be commonly used and too slow.
*/

unsigned int Cor_Relation::cardinality(Cor_Tuple *T)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_sets) {
            *(Cor_exEnv->trace_strm) << 
                "Relation::cardinality ----- cardinality of:  ";
            T->printon(Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << endl;
        }
#    endif

#ifdef    DEBUG
    HashVal h_T = T->hash();
    ASSERT((Cor_isGroundHashValue(h_T)));
#endif

    Cor_Tuple *tuple;
    HashVal h_tuple;
    unsigned int count = 0;

    FOR_EACH_TUPLE(tuple, this) {
        h_tuple = tuple->hash();
        ASSERT((Cor_isGroundHashValue(h_tuple)));

        if (Cor_equal_tuples(tuple, T, /*display=*/ 0))
            count++;
    } END_EACH_TUPLE(tuple)

#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_sets) {
            *(Cor_exEnv->trace_strm) << 
                "            cardinality is: " <<  count << ".\n";
        }
#    endif

    return(count);
}

// *************************************************************************
//    return true iff (*this) is a subset of rel

int Cor_Relation::contained_in(Cor_Relation *rel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_sets)
        *(Cor_exEnv->trace_strm) << 
            "Relation::contained_in ----- check set containment.\n";
#    endif

    Cor_Tuple *tuple;
    int contain = 1;

    FOR_EACH_TUPLE(tuple,this) {
        if (this->cardinality(tuple) > rel->cardinality(tuple)) {
            contain = 0;
            break;
        }
    } END_EACH_TUPLE(tuple)

#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_sets) {
            if (contain) {
                *(Cor_exEnv->trace_strm) << 
                    "       ----- Relation::contained_in SUCCEEDED.\n";
            }
            else {
                *(Cor_exEnv->trace_strm) << 
                    "       ----- Relation::contained_in FAILED.\n";
            }
        }
#    endif

    return(contain);
}

// *************************************************************************
/*
    Return true iff *rel = *this.
    Inputs can be assumed to be ground, since this is only called with
    nested relations for which this condition is satisfied.
*/

int Cor_Relation::raw_equals(Cor_Relation *rel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_sets) {
            *(Cor_exEnv->trace_strm) << 
                "Relation::raw_equals ----- executing set equality.\n";
        }
#    endif

    Cor_Tuple *tuple = NULL;
    int equals;

    // first check that the two relations have the same cardinality
    // or else do not proceed.
    if (this->count != rel->count)
        equals = 0;
    else {
        equals = 1;
        FOR_EACH_TUPLE(tuple,this) {
            if (this->cardinality(tuple) != rel->cardinality(tuple)) {
                equals = 0;
                break;
            }
        } END_EACH_TUPLE(tuple)
    }

#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_sets) {
            if (equals) {
                *(Cor_exEnv->trace_strm) << 
                    "        ----- Relation::raw_equals SUCCEEDED.\n";
            }
            else {
                *(Cor_exEnv->trace_strm) << 
                    "        ----- Relation::raw_equals FAILED.\n";
            }
        }
#    endif

    return equals;
}

// *************************************************************************

// Make sure this is not equal to UnknownHashValue and VarHashValue
static int C_eq_id_count = 1;

int Cor_Relation::equals(Cor_Relation *that)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_sets) {
            *(Cor_exEnv->trace_strm) << 
                "Relation::equals ------ executing set equality.\n";
        }
#    endif

    HashVal h1 = this->hash();
    HashVal h2 = that->hash();

#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_sets) {
            *(Cor_exEnv->trace_strm) << "Set-1 = ";
            printon(Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << ". Hash = "
                << hex << h1 << dec << endl << endl;
            *(Cor_exEnv->trace_strm) << "Set-2 = ";
            that->printon(Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << ". Hash = "
                << hex << h2 << dec << endl << endl;
        }
#    endif

    if ( !Cor_isGroundHashValue(h1) || !Cor_isGroundHashValue(h2) ) {
        CORAL_error(COR_NESTED_SET_NOT_GROUND, NULL, "Relation::equals");
        return(-1);
    }

    if ( h1 != h2 ) {
#        ifdef DO_TRACE
            if (Cor_exEnv->dbg_sets)
            *(Cor_exEnv->trace_strm) << 
                    "          ------- set equality FAILED.\n";
#        endif
        return 0;
    }

    if ( Cor_isGroundHashValue(_eq_id) && (_eq_id == that->_eq_id) ) {
#        ifdef DO_TRACE
            if (Cor_exEnv->dbg_sets) {
                *(Cor_exEnv->trace_strm) << 
                        "          ground id values exist.\n";
                *(Cor_exEnv->trace_strm) << 
                    "          ------- set equality SUCCEEDED.\n";
            }
#        endif
        return 1;
    }

    int rel_equals = raw_equals(that);

    if ( !rel_equals) {
#        ifdef DO_TRACE
            if (Cor_exEnv->dbg_sets) {
                *(Cor_exEnv->trace_strm) << 
                    "          ------- set equality FAILED.\n";
            }
#        endif
        return 0;
    }

    // The sets are equal, so assign them the same id

    if (Cor_isGroundHashValue(_eq_id) && Cor_isGroundHashValue(that->_eq_id)) {

        HashVal min_val = (that->_eq_id < _eq_id)? that->_eq_id : _eq_id;
        _eq_id = that->_eq_id = min_val;
    } else if ( ( ! Cor_isGroundHashValue(_eq_id))
           && ( ! Cor_isGroundHashValue(that->_eq_id) ) ) {
        _eq_id = that->_eq_id = C_eq_id_count++;
    } else {
        // exactly one of them has been previously assigned
        int id_val = Cor_isGroundHashValue(_eq_id) ? _eq_id : that->_eq_id;
        _eq_id = that->_eq_id = id_val;
    }

#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_sets) {
            *(Cor_exEnv->trace_strm) << 
                "           assigning equal id values: " <<  _eq_id << ".\n";
            *(Cor_exEnv->trace_strm) << 
                "           ------- set equality SUCCEEDED.\n";
        }
#    endif

    return 1;
}

#ifdef ASYN_SOCK_REL
// *************************************************************************

void Cor_Relation::send_delta(Cor_QFMaterializedModuleData *md,
        Cor_Table *table, int scc_id, int pred_num, Cor_Predicate *pred)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *base, *magic=NULL, *adorn = NULL;

    if (first_send_delta) {

        // pred's name is of the form m_<name> or m_<name>_[bf]*
        char tmp1[100], tmp3[100];
        strcpy(tmp1, pred->name.symbol());

        base = Cor_split_relname(tmp1, &magic, &adorn);

        if (adorn && *adorn)
          sprintf(tmp3,"%s_%s",base,adorn);
        else
          strcpy(tmp3,base);

        // rel is the (storage) asyn_sock_rel
        Cor_Relation *rel
            = Cor_find_external_relation(Cor_CArg(base), pred->arity);
        ASSERT(rel);
        
        // rel_ext is the (external) asyn_sock_rel
        Cor_Relation *rel_ext
            = Cor_find_local_relation(Cor_CArg(tmp3),rel->_arity);

        if (!rel_ext) {
            cerr << "Could not locate relation: " << tmp3 << endl;
            return;
        }

        magic_mark[0] = NULL;        // set initial marks
        magic_mark[1] = NULL;

        cs = rel_ext->cs = ((Cor_AsynSockRelation*)rel)->setup();
    }

    if (!cs)
        return;

    freeMark(magic_mark[0]);
    magic_mark[0] = magic_mark[1];
    magic_mark[1] = getMark();

#ifdef DEBUG_ASYN_SOCK_REL
    cout << "sending query based on delta for " << pred->name << endl;
#endif
    
    Cor_ArgList *al = make_vararglist(pred->arity);
    Cor_BindEnv *env = new Cor_BindEnv(pred->arity);
    Cor_RMark   *start_mark, *end_mark;

    start_mark = magic_mark[0];
    end_mark   = magic_mark[1];

    Cor_TupleIterator *it;

    it = new Cor_TupleIterator(this,*al,env, NULL, start_mark, end_mark);
    it->reset();

    char tmp1[100],tmp3[100];

    strcpy(tmp1,pred->name.symbol());
    base = Cor_split_relname(tmp1, &magic, &adorn);

    if (adorn && *adorn)
        sprintf(tmp3,"%s_%s",base,adorn);
    else
        strcpy(tmp3,base);

    Cor_Relation *rel_ext =
        (Cor_Relation *) Cor_CurDB->RelationTable->lookup(Cor_CArg(base));
    ASSERT(rel_ext);
    Cor_AsynSockRelation *srel = (Cor_AsynSockRelation*) rel_ext;
    int len = rel_ext->_arity;
    Cor_Relation *rel = Cor_find_local_relation(Cor_CArg(tmp3),rel_ext->_arity);

    if (!rel) {
        cerr << "Could not find relation " << tmp3 << endl;
        return;
    }

    char query_buf[1000];
    for ( Cor_Tuple *tuple = it->get_next_tuple();
         !it->no_match(); tuple = it->get_next_tuple()) {

        if (tuple) {

            // if pred is the magic set of the b..b adornment of the
            // asyn_sock_rel. the tuples in that magic set won't change
            // after each iteration. Hence we have to ignore it for the
            // remaining time. 

            if (!first_send_delta && (!adorn || !strchr(adorn,'f')))
                break;

            // construct queries sent to the server based
            // on new facts in magic relation
            rel->_delta_done = 0;       // have more results coming anyway
            ostrstream query(query_buf, 1000);

            if (!adorn) {
                ASSERT(srel->remote_relname());
                query << "?" << srel->remote_relname() << "(";
                for (int i = 0; i < len; i++) {
                    query << "X" << i;
                    if (i != len-1)
                        query << ",";
                    else
                        query << ")";
                }
            } else {
                query << "?" << srel->remote_relname() << "(";
                for (int i = 0, j=0; i < len; i++){
                    if (adorn[i] == 'b') // bound
                        query << tuple->_args[j++];
                    else // free
                        query << "X" << i;
                    if (i != len-1)
                        query << ",";
                    else
                        query << ")";
                }
            }
            query << '\0';
            
            *cs << query_buf << endl << flush;

            rel->open_queries++;
 
#ifdef DEBUG_ASYN_SOCK_REL
            cout << "tuple = (" << *tuple << ").\n";
            cout << "query = " << query_buf << endl;
#endif
        }
    }

    delete it;

    first_send_delta = 0;
}
#endif // ASYN_SOCK_REL
// *************************************************************************
